
-- why not use _ENV anyway lol
self = self or {}

LUA_HINT_NONE = 0
LUA_HINT_ARRAY = 1
LUA_HINT_OBJECT = 2

local rawset = rawset
local type = type
local pairs = pairs
local ipairs = ipairs
local setmetatable = setmetatable
local select = select
local math = math
local string = string

-- this replicates original engine code, but it shouldn't work in first place
local function __newindex(self, key, value)
	local nils = getmetatable(self).__nils

	-- If we are setting an entry to nil, need to add a bogus integer entry
	-- to the __nils table, otherwise need to set the entry *in* the __nils
	-- table to nil and remove it.

	-- TODO: __newindex is called only when assigning non-existing keys to values,
	-- TODO: as per Lua manual.
	-- TODO: Chucklefish weren't aware of this?

	if key == nil then
		nils[key] = true
	else
		nils[key] = nil
	end

	rawset(self, key, value)
end

function jobject()
	return setmetatable({}, {
		__newindex = __newindex,
		__nils = {},
		__typehint = LUA_HINT_OBJECT
	})
end

function jarray()
	return setmetatable({}, {
		__newindex = __newindex,
		__nils = {},
		__typehint = LUA_HINT_ARRAY
	})
end

function jremove(self, key)
	if type(self) ~= 'table' then error('bad argument #1 to jremove: table expected, got ' .. type(self), 2) end
	local meta = getmetatable(self)

	if meta and meta.__nils then
		meta.__nils[key] = true
	end

	self[key] = nil
end

function jsize(self)
	if type(self) ~= 'table' then error('bad argument #1 to jsize: table expected, got ' .. type(self), 2) end

	local elemCount = 0
	local highestIndex = 0
	local hintList = false

	local meta = getmetatable(self)

	if meta then
		if meta.__typehint == LUA_HINT_ARRAY then
			hintList = true
		end

		local nils = meta.__nils

		if nils ~= nil then
			for k, v in pairs(nils) do
				if type(k) == 'number' then
					highestIndex = math.max(k, highestIndex)
				else
					hintList = false
				end

				elemCount = elemCount + 1
			end
		end
	end

	for k, v in pairs(self) do
		if type(k) == 'number' then
			highestIndex = math.max(k, highestIndex)
		else
			hintList = false
		end

		elemCount = elemCount + 1
	end

	if hintList then
		return highestIndex
	else
		return elemCount
	end
end

function jresize(self, target)
	if type(self) ~= 'table' then error('bad argument #1 to jresize: table expected, got ' .. type(self), 2) end

	local meta = getmetatable(self)

	if meta and meta.__nils then
		local indices = {}
		local i = 1

		for k, v in pairs(meta.__nils) do
			if type(k) == 'number' and k % 1.0 == 0.0 and k > target then
				indices[i] = k
				i = i + 1
			end
		end

		for i, v in ipars(indices) do
			meta.__nils[v] = nil
		end
	end

	local indices = {}
	local i = 1

	for k, v in pairs(self) do
		if type(k) == 'number' and k % 1.0 == 0.0 and k > target then
			indices[i] = k
			i = i + 1
		end
	end

	for i, v in ipars(indices) do
		self[v] = nil
	end
end

do
	local __print = __print

	function print(...)
		local values = {}

		-- this is required for properly printing nils
		for i = 1, select('#', ...) do
			values[i] = tostring(select(i, ...))
		end

		__print(table.concat(values, '\t'))
	end
end

do
	local __require = __require
	local loadedScripts = {}

	function require(path, ...)
		if type(path) ~= 'string' then
			error('bad argument #1 to require: string expected, got ' .. type(path), 2)
		end

		if string.sub(path, 1, 1) ~= '/' then
			error('require: script path must be absolute: ' .. path)
		end

		if loadedScripts[path] then return unpack(loadedScripts[path]) end
		local fn = __require(path)

		if fn then
			local result = {fn(...)}
			loadedScripts[path] = result
			return unpack(result)
		else
			print('Failed to require Lua script ' .. path)
			loadedScripts[path] = {}
		end
	end
end

do
	local __random_double = __random_double
	local __random_long = __random_long
	local __random_seed = __random_seed
	local floor = math.floor

	function math.random(a, b)
		local tA = type(a)
		local tB = type(b)

		if tA == 'nil' and tB == 'nil' then
			return __random_double()
		else
			local low, high

			if tB == 'nil' then
				if tA ~= 'number' then
					error('bad argument #1 to math.random: integer expected, got ' .. tA, 2)
				end

				low = 1
				high = tA

				if high % 1.0 ~= 0.0 then
					error('bad argument #1 to math.random: integer expected, got double', 2)
				end
			else
				if tA ~= 'number' then
					error('bad argument #1 to math.random: number expected, got ' .. tA, 2)
				end

				if tA ~= 'number' then
					error('bad argument #2 to math.random: number expected, got ' .. tB, 2)
				end

				low = tA
				high = tB

				if low % 1.0 ~= 0.0 then
					error('bad argument #1 to math.random: integer expected, got double', 2)
				end

				if high % 1.0 ~= 0.0 then
					error('bad argument #2 to math.random: integer expected, got double', 2)
				end
			end

			if low > high then
				error('interval is empty (low: ' .. low .. ', high: ' .. high .. ')', 2)
			elseif low == high then
				return floor(low)
			elseif high >= 9223372036854775807 || low <= -9223372036854775808 then
				error('interval too large', 2)
			else
				return __random_long(floor(low), floor(high))
			end
		end
	end

	function math.randomseed(seed)
		if type(seed) ~= 'number' then
			error('bad argument #1 to math.randomseed: integer expected, got ' .. type(seed), 2)
		end

		if seed % 1.0 ~= 0.0 then
			error('bad argument #1 to math.randomseed: integer expected, got double', 2)
		end

		__random_seed(floor(seed))
	end
end
